
/*******************************************************************************

 Copyright (c) 2009, Charles McGarvey
 All rights reserved.
 
 Redistribution   and   use  in  source  and  binary  forms,  with  or  without
 modification, are permitted provided that the following conditions are met:
 
   * Redistributions  of  source  code  must retain the above copyright notice,
     this list of conditions and the following disclaimer.
   * Redistributions  in binary form must reproduce the above copyright notice,
     this  list of conditions and the following disclaimer in the documentation
     and/or other materials provided with the distribution.
 
 THIS  SOFTWARE  IS  PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 AND  ANY  EXPRESS  OR  IMPLIED  WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 DISCLAIMED.  IN  NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
 FOR  ANY  DIRECT,  INDIRECT,  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 DAMAGES  (INCLUDING,  BUT  NOT  LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 SERVICES;  LOSS  OF  USE,  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 CAUSED  AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 OR  TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

*******************************************************************************/

#include <cstdlib>		// atexit, getenv
#include <iostream>
#include <string>
#include <unistd.h>		// access

#include <Moof/Dispatcher.hh>
#include <Moof/Exception.hh>
#include <Moof/Log.hh>
#include <Moof/OpenGL.hh>
#include <Moof/Resource.hh>
#include <Moof/Transition.hh>
#include <Moof/Video.hh>

#include "GameLayer.hh"
#include "MainLayer.hh"
#include "TitleLayer.hh"

#if HAVE_CONFIG_H
#include "config.h"
#endif
#include "version.h"


MainLayer::MainLayer()
{
	Mf::dispatcher::addHandler("video.context_recreated",
			boost::bind(&MainLayer::contextRecreated, this, _1), this);
	setupGL();
}

MainLayer::~MainLayer()
{
	Mf::dispatcher::removeHandler(this);
}


void MainLayer::pushed(Mf::Engine& e)
{
	engine = &e;

	//Mf::Scalar coeff[] = {0.0, 1.0};
	//Mf::Lerp interp(coeff, 0.25);

	//Mf::LayerP gameLayer = GameLayer::alloc();
	//Mf::Transition<Mf::Lerp>::Ptr transition =
		//Mf::Transition<Mf::Lerp>::alloc(gameLayer, Mf::LayerP(), interp);
	//engine->push(transition);
	//engine->push(GameLayer::alloc());
	engine->push(TitleLayer::alloc());
}


void MainLayer::draw(Mf::Scalar alpha) const
{
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();

	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
}

bool MainLayer::handleEvent(const Mf::Event& event)
{
	switch (event.type)
	{
		case SDL_KEYDOWN:
			if (event.key.keysym.sym == SDLK_ESCAPE)
			{
				quit();
			}
			else if (event.key.keysym.sym == SDLK_f)
			{
				engine->getVideo().toggleFull();
			}
			else if (event.key.keysym.sym == SDLK_l)
			{
				Mf::Video& video = engine->getVideo();
				video.toggleCursorGrab();
				video.toggleCursorVisible();
			}
			else if (event.key.keysym.sym == SDLK_y)
			{
				engine->push(GameLayer::alloc());
			}
			break;

		case SDL_VIDEORESIZE:
			glViewport(0, 0, event.resize.w, event.resize.h);
			break;

		case SDL_QUIT:
			quit();
			break;
	}

	return false;
}

void MainLayer::quit()
{
#if NDEBUG
	// we don't really need to unwind the stack and shut everything down because
	// the operating system will take care of cleaning up
	exit(0);
#else
	engine->clear();
#endif
}


void MainLayer::setupGL()
{
	glEnable(GL_TEXTURE_2D);
	glEnable(GL_DEPTH_TEST);

	glEnable(GL_LINE_SMOOTH);
	glEnable(GL_POLYGON_SMOOTH);
	glShadeModel(GL_SMOOTH);

	//glEnable(GL_BLEND);
	//glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	glEnable(GL_ALPHA_TEST);
	glAlphaFunc(GL_GREATER, 0.0);

	glClearColor(0.0, 0.0, 0.0, 1.0);

	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	gluPerspective(60.0, 1.33333, 1.0, 2500.0);

	glMatrixMode(GL_MODELVIEW);
}

void MainLayer::contextRecreated(const Mf::Notification* note)
{
	// whenever the context is destroyed and a new one created, it probably
	// won't contain our state so we need to set that up again
	setupGL();
}



void printUsage()
{
	std::cout << "Usage: "PACKAGE" [-h|--help] [-i|--info] [OPTION=VALUE]..." << std::endl
			  << "The alien-smashing action game." << std::endl
			  << std::endl
			  << "Options:" << std::endl
			  << "  -h, --help" << std::endl
			  << "      show this help and exit" << std::endl
			  << "  -i, --info" << std::endl
			  << "      show version and build information" << std::endl
			  << "  detail=1|2|3" << std::endl
			  << "      the level of detail of game scenes" << std::endl
			  << "  fullscreen=true|false" << std::endl
			  << "      if true, uses the entire display" << std::endl
			  << "  maxfps=num" << std::endl
			  << "      the maximum number of frames per second" << std::endl
			  << std::endl
			  << "See documentation for more options." << std::endl;
}

void printInfo(int argc, char* argv[])
{
	std::string assets;
	std::string datadir;
	std::string config;

	assets.assign(YOINK_DATADIR);
	int accessible = access(assets.c_str(), R_OK);
	if (accessible != 0) assets += " (no access)";

	char* temp = getenv("YOINK_DATADIR");
	if (temp)
	{
		datadir = temp;
		accessible = access(temp, R_OK);
		if (accessible != 0) datadir += " (no access)";
	}

	temp = getenv("YOINKRC");
	if (temp)
	{
		config = temp;
		accessible = access(temp, R_OK);
		if (accessible != 0) config += " (no access)";
	}

	std::cout << "   Executable: " << argv[0] << std::endl
			  << "      Version: "VERSION << std::endl
#ifdef __DATE__
			  << "        Built: "__DATE__" "__TIME__ << std::endl
#endif
			  << "     Compiler: "COMPILER_STRING << std::endl
			  << "       Assets: " << assets << std::endl
			  << "Build options: "
#ifdef NDEBUG
			  << "-"
#endif
			  << "debug "
#ifndef USE_DOUBLE_PRECISION
			  << "-"
#endif
			  << "double-precision "
#ifndef PROFILING_ENABLED
			  << "-"
#endif
			  << "profile "
#ifndef USE_THREADS
			  << "-"
#endif
			  << "threads" << std::endl
			  << "      YOINKRC: " << config << std::endl
			  << "YOINK_DATADIR: " << datadir << std::endl;
}

void goodbye()
{
	std::cout << std::endl << "Goodbye..." << std::endl << std::endl;
}



int main(int argc, char* argv[])
{
	if (argc > 1)
	{
		std::string arg(argv[1]);
		if (arg == "-h" || arg == "--help")
		{
			printUsage();
			return 0;
		}
		else if (arg == "-i" || arg == "--info")
		{
			printInfo(argc, argv);
			return 0;
		}
	}

	std::cout << std::endl << PACKAGE_STRING << std::endl
			  << "Compiled " << __TIME__ " " __DATE__ << std::endl
			  << "Send patches and bug reports to <"
			  PACKAGE_BUGREPORT << ">." << std::endl << std::endl;

	atexit(goodbye);


#if		YOINK_LOGLEVEL >= 4
	Mf::setLogLevel(Mf::LOG_DEBUG);
#elif	YOINK_LOGLEVEL >= 3
	Mf::setLogLevel(Mf::LOG_INFO);
#elif	YOINK_LOGLEVEL >= 2
	Mf::setLogLevel(Mf::LOG_SCRIPT);
#elif	YOINK_LOGLEVEL >= 1
	Mf::setLogLevel(Mf::LOG_ERROR);
#elif	YOINK_LOGLEVEL
	Mf::setLogLevel(Mf::LOG_NONE);
#endif


	// Add search paths; they should be searched in this order:
	// 1. YOINK_DATADIR (environment)
	// 2. YOINK_DATADIR (configure)

	char* dataDir = getenv("YOINK_DATADIR");
	if (dataDir)
	{
		Mf::Resource::addSearchPath(dataDir);
	}

	Mf::Resource::addSearchPath(YOINK_DATADIR);

	std::string iconFile = Mf::Resource::getPath("yoink.png");


	// Build the list of config files to search for, in this order:
	// 1. YOINK_DATADIR/yoinkrc
	// 2. /etc/yoinkrc
	// 3. $HOME/.yoinkrc
	// 4. YOINKRC (environment)

	std::string configFiles;

	configFiles += Mf::Resource::getPath("yoinkrc");
	configFiles += ":/etc/yoinkrc:$HOME/.yoinkrc";

	char* configFile = getenv("YOINKRC");
	if (configFile)
	{
		configFiles += ":";
		configFiles += configFile;
	}


	try
	{
		Mf::Engine app(argc, argv, PACKAGE_STRING, iconFile, configFiles);
		app.push(MainLayer::alloc());

		app.run();
	}
	catch (Mf::Exception e)
	{
		Mf::logError("unhandled exception: <<%s>>", e.what());
		Mf::logInfo("it's time to crash now :-(");
		throw e;
	}

	return 0;
}


/** vim: set ts=4 sw=4 tw=80: *************************************************/

